/*
Author: Jimmy Reumuth
Date: Oct 19, 2001

Pointer is a smart pointer, i.e. a pointer that keeps track of
all references to its allocated data and deletes it when no more
refernces exist. It should be impossible to have memory leaks using
Pointer.

Note there are three ways for a pointer to iterate through its
elements:
for (Pointer<T> it = pointer; it < pointer + pointer.Size(); ++it)
	//do stuff with it-> or *it

for (Pointer<T> it = pointer; it.Offset() < pointer.Size(); ++it)
	//do stuff with it-> or *it

for (i = 0; i < pointer.Size(); ++i)
	//do stuff with pointer[i]

The first one is terribly inefficient. Every evaluation of
pointer + pointer.Size() calls a constructor, destructor, and +=.
*/

#ifndef _POINTER_H
#define _POINTER_H

#include "Functional.h"
#include <cassert>

//Warning for operator-> on non struct or class types, e.g.
//Pointer<int> pInt; pInt->
//If someone tries to do that it won't compile anyway.
#pragma warning (disable : 4284)

#ifndef NULL
#define NULL 0
#endif

template<class T>
class Pointer : public Eq<Pointer>, public Ord<Pointer> {
public:
	typedef unsigned size_type;

private:
	T *m_pData;

	//Count of number of refernces to m_pData. Copy constructor and
	//operator= ++ it, and destructor and operator= -- it.
	unsigned *m_pnRefCount;

	//m_nOffset is for "moving" the pointer with +, ++, etc. operators
	size_type m_nOffset, *m_pnSize;

	//Copy makes this point to the same data as rhs.
	void Copy(const Pointer &rhs) {
		m_pData = rhs.m_pData;
		m_pnSize = rhs.m_pnSize;
		++*(m_pnRefCount = rhs.m_pnRefCount);
		m_nOffset = rhs.m_nOffset;
	}

	void ReleaseReference() {
		if (*m_pnRefCount == 0) {
			delete m_pnRefCount;
			delete m_pnSize;
			delete [] m_pData;
		} else
			--*m_pnRefCount;
	}

public:
	Pointer(size_type size = 0) :
	m_pData(size == 0 ? NULL : new T[size]),
	m_pnRefCount(new unsigned),
	m_pnSize(new size_type),
	m_nOffset(0) {
		*m_pnSize = size;
		*m_pnRefCount = 0;
	}

	Pointer(const Pointer &rhs) {Copy(rhs);}
	~Pointer() {ReleaseReference();}

	//Allocates new data for this which is given the same values
	//as rhs's data. The data in rhs is copied starting at 
	//rhs.Offset().
	void DeepCopy(const Pointer &rhs) {
		ReleaseReference();

		//Don't need to delete old data. ReleaseReference() takes
		//care of that.
		*(m_pnRefCount = new unsigned) = 0;
		m_pnSize = new size_type;
		if ((*m_pnSize = *rhs.m_pnSize - rhs.m_nOffset) <= 0) {
			*m_pnSize = 0;
			m_pData = NULL;
			return;
		}

		m_pData = new T[*m_pnSize];
		for (size_type i = 0; i < *m_pnSize; ++i)
			m_pData[i] = rhs.m_pData[rhs.m_nOffset + i];
	}

	size_type Offset() const {return m_nOffset;}

	//No assertsions here. It's perfectly valid to have a pointer
	//go out of range, as long as you don't dereference it while
	//it's out of range.
	T* Ptr() {return m_pData + m_nOffset;}
	const T* Ptr() const {return m_pData + m_nOffset;}

	size_type Size() const {return *m_pnSize;}

	Pointer& operator=(const Pointer& rhs) {
		ReleaseReference();
		Copy(rhs);
		return *this;
	}

	bool operator==(const Pointer& rhs) const {
		return m_pData == rhs.m_pData && m_nOffset == rhs.m_nOffset;
	}

	bool operator<(const Pointer& rhs) const {
		return m_pData + m_nOffset < rhs.m_pData + rhs.m_nOffset;
	}

	//pre-incriment
	Pointer& operator++() {
		++m_nOffset;
		return *this;
	}

	//post-incriment
	const Pointer operator++(int) {
		Pointer temp = *this;
		++*this;
		return temp;
	}

	//pre-decriment
	Pointer& operator--() {
		++m_nOffset;
		return *this;
	}

	//post-decriment
	const Pointer operator--(int) {
		Pointer temp = *this;
		--*this;
		return temp;
	}

	Pointer& operator+=(size_type inc_by) {
		m_nOffset += inc_by;
		return *this;
	}

	const Pointer operator+(size_type inc_by) const {
		return Pointer(*this) += inc_by;
	}


	Pointer& operator-=(size_type dec_by) {
		m_nOffset -= dec_by;
		return *this;
	}

	const Pointer operator-(size_type dec_by) const {
		return Pointer(*this) -= dec_by;
	}

	T& operator*() {
		assert(m_pData != NULL);
		assert(m_nOffset < *m_pnSize);
		return m_pData[m_nOffset];
	}

	const T& operator*() const {
		assert(m_pData != NULL);
		assert(m_nOffset < *m_pnSize);
		return m_pData[m_nOffset];
	}

	T* operator->() const {
		assert(m_pData != NULL);
		assert(m_nOffset < *m_pnSize);
		return m_pData + m_nOffset;
	}

	T& operator[](size_type offset) {
		assert(m_pData != NULL);
		assert(m_nOffset + offset < *m_pnSize);
		return m_pData[m_nOffset + offset];
	}

	const T& operator[](size_type offset) const {
		assert(m_pData != NULL);
		assert(m_nOffset + offset < *m_pnSize);
		return m_pData[m_nOffset + offset];
	}
};

#endif